from typing import List

pseudo: List[str] = [
    "aucun changement",
    "aucun doute",
    "aucune hésitation",
    "aucune diminution",
    "ne permet pas d",
    "ne permet pas de",
    "n'exclut pas",
    "non négligeable",
    "pas d'amélioration",
    "pas d'augmentation",
    "pas d'autre",
    "pas de changement",
    "pas de diminution",
    "pas de doute",
    "pas exclu",
    "pas exclue",
    "pas exclues",
    "pas exclus",
    "pas immunisé",
    "pas immunisée",
    "pas immunisés",
    "pas immunisées",
    "sans amélioration",
    "sans aucun doute",
    "sans augmentation",
    "sans certitude",
    "sans changement",
    "sans diminution",
    "sans doute",
    "sans être certain",
]

preceding: List[str] = [
    "à la place de",
    "absence",
    "absence de signe de",
    "absence de",
    "aucun signe de",
    "aucun",
    "aucune preuve",
    "aucune",
    "aucunes",
    "aucuns",
    "décline",
    "décliné",
    "dépourvu",
    "dépourvue",
    "dépourvues",
    "dépourvus",
    "disparition de",
    "disparition des",
    "excluent",
    "exclut",
    "impossibilité de",
    "immunisé",
    "immunisée",
    "immunisés",
    "immunisées",
    "incompatible avec",
    "incompatibles avec",
    "jamais",
    "ne manifestaient pas",
    "ne manifestait pas",
    "ne manifeste pas",
    "ne manifestent pas",
    "ne pas",
    "ne présentaient pas",
    "ne présentait pas",
    "ne présente pas",
    "ne présentent pas",
    "ne ressemble pas",
    "ne ressemblent pas",
    "négatif pour",
    "n'est pas",
    "n'était pas",
    "ni",
    "niant",
    "nie",
    "nié",
    "non",
    "nullement",
    "pas d",
    "pas de cause de",
    "pas de signe de",
    "pas de signes de",
    "pas de",
    "pas nécessaire de",
    "pas",
    "permet d'exclure",
    "plus d'aspect de",
    "sans manifester de",
    "sans présenter de",
    "sans",
    "symptôme atypique",
]

following: List[str] = [
    ":0",
    ": 0",
    ":non",
    ": non",
    "absent",
    "absente",
    "absentes",
    "absents",
    "dépourvu",
    "dépourvue",
    "dépourvues",
    "dépourvus",
    "disparaissent",
    "disparait",
    "est exclu",
    "est exclue",
    "immunisé",
    "immunisée",
    "immunisés",
    "immunisées",
    "impossible",
    "improbable",
    "négatif",
    "négatifs",
    "négative",
    "négatives",
    "négligeable",
    "négligeables",
    "nié",
    "niée",
    "non",
    "pas nécessaire",
    "peu probable",
    "sont exclues",
    "sont exclus",
]

verbs: List[str] = [
    "éliminer",
    "exclure",
    "interdire",
    "nier",
    "réfuter",
    "rejeter",
]
