from typing import List

family: List[str] = [
    "aïeul",
    "aïeux",
    "antécédent familial",
    "antécédents familiaux",
    "arrière-grand-mère",
    "arrière-grand-père",
    "arrière-grands-parents",
    "cousin",
    "cousine",
    "cousines",
    "cousins",
    "enfant",
    "enfants",
    "épouse",
    "époux",
    "familial",
    "familiale",
    "familiales",
    "familiaux",
    "famille",
    "fiancé",
    "fiancée",
    "fils",
    "frère",
    "frères",
    "grand-mère",
    "grand-père",
    "grands-parents",
    "maman",
    "mari",
    "mère",
    "oncle",
    "papa",
    "parent",
    "parents",
    "père",
    "soeur",
    "sœur",
    "sœurs",
    "soeurs",
    "tante",
]
