consultation_mention = [
    "rendez-vous pris",
    r"consultation",
    r"consultation.{1,8}examen",
    r"\bcs\b",
    "examen clinique",
    r"de compte rendu",
    r"date de l'examen",
    r"examen realise le",
    "date de la visite",
]

town_mention = [
    "paris",
    "kremlin.bicetre",
    "creteil",
    "boulogne.billancourt",
    "villejuif",
    "clamart",
    "bobigny",
    "clichy",
    "ivry.sur.seine",
    "issy.les.moulineaux",
    "draveil",
    "limeil",
    "champcueil",
    "roche.guyon",
    "bondy",
    "colombes",
    "hendaye",
    "herck.sur.mer",
    "labruyere",
    "garches",
    "sevran",
    "hyeres",
]

document_date_mention = [
    "imprime le",
    r"signe electroniquement",
    "signe le",
    "saisi le",
    "dicte le",
    "tape le",
    "date de reference",
    r"date\s*:",
    "dactylographie le",
    "date du rapport",
]
